import asyncio
from asyncio.futures import Future
from qtpy import QtGui
from qtpy.QtCore import QObject, Signal
import uuid
import logging
from koil.koil import Koil, get_current_koil

logger = logging.getLogger(__name__)


Reference = str

class UnconnectedSignalError(Exception):
    pass


def get_receiver_length(qobject, qsignal, callstring):
    try:
        return qobject.receivers(qsignal)
    except:
        return qobject.receivers(callstring)





class FutureWrapper(QObject):
    call = Signal(Reference, tuple, dict)
    resolve = Signal(Reference, object)
    reject = Signal(Reference, Exception)
    cancel = Signal(Reference, tuple, dict)
    cancelled = Signal(Reference)

    
    def __init__(self, *args, koil: Koil = None, cancel_timeout = 4, pass_through=False, **kwargs) -> None:
        super().__init__( *args, **kwargs)
        self.resolve.connect(self.on_resolve)
        self.reject.connect(self.on_reject)
        self.cancelled.connect(self.on_cancelled)
        self.futureMap = {}
        self.cancelMap = {}
        self.koil = koil or get_current_koil()
        self.loop = self.koil.loop
        self.pass_through = pass_through
        self.cancel_timeout = 4


    def on_cancelled(self, ref):
        assert ref in self.cancelMap, "Received cancellation for non invoked cancel future"
        cancel_future = self.cancelMap[ref]
        if not cancel_future.done():
            self.loop.call_soon_threadsafe(cancel_future.set_result, None)


    def on_resolve(self, ref, obj):
        assert ref in self.futureMap, "Received resolvation for non invoked future"
        future = self.futureMap[ref]
        if not future.done():
            self.loop.call_soon_threadsafe(future.set_result, obj)


    def on_reject(self, ref, exp):
        assert ref in self.futureMap, "Received rejection for non invoked future"
        future = self.futureMap[ref]
        if not future.done():
            self.loop.call_soon_threadsafe(future.set_exception, exp)



    async def acall(self, *args, **kwargs):

        if get_receiver_length(self, self.call, "call(str, tuple, dict)") == 0:
           if self.pass_through: return None
           else: raise UnconnectedSignalError("This future has no connected receivers")


        
        reference = str(uuid.uuid4())
        future = self.loop.create_future()
        self.futureMap[reference] = future
        self.call.emit(reference, args, kwargs)
        try:
            return await future
        except asyncio.CancelledError as e:
            if self.receivers(self.cancel) > 0:
                cancel_future = self.loop.create_future()
                self.cancelMap[reference] = future
                self.cancel.emit(reference, args, kwargs)
                try:
                    await asyncio.wait_for(cancel_future, timeout=self.cancel_timeout)
                    raise e
                except asyncio.TimeoutError as e:
                    logger.error("Cancellation recevied timeout, Cancelling anyways")
                    raise e
            else: 
                raise e






class TimeoutFutureWrapper(FutureWrapper):

    def __init__(self, *args, timeout=None, **kwargs) -> None:
        super().__init__( *args, **kwargs)
        self.timeout = timeout

    async def acall(self, *args, **kwargs):
        return await asyncio.wait_for(super().acall(*args, **kwargs), timeout=self.timeout)




import logging
import asyncio




logger = logging.getLogger(__name__)


class KoilTask:

    def __init__(self, future=None, koil = None, *args, log_errors=True, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        from koil.koil import get_current_koil
        self.future = future
        self.task = None
        self.log_errors = log_errors
        self.koil = koil or get_current_koil()
        self.loop = self.koil.loop
        self.loop.call_soon_threadsafe(self.run)

    async def wrapped_future(self, future):
        try:
            return await future
        except Exception as e:
            if self.log_errors: logger.exception(e)
            raise e


    def run(self):
        self.task = self.loop.create_task(self.wrapped_future(self.future))

    async def acancel(self):
        try:
            self.task.cancel()
            try:
                await self.task
            except asyncio.CancelledError as e:
                logger.debug("Cancelled KoilTask")
        except Exception as e:
            logger.error("Koil Task Cancellation failed")

    def cancel(self):
        return asyncio.run_coroutine_threadsafe(self.acancel(), self.loop).result()





from abc import abstractmethod
import abc
from typing import Type, Union
from koil.exceptions import KoilException
from koil.state import KoilState





class CheckerExpection(KoilException):
    pass


class BaseChecker(abc.ABC):
    pass

    def __init__(self, koil) -> None:
        self.koil = koil
        super().__init__()


    @abstractmethod
    def force_state(self) -> Union[None, KoilState]:
        """Force State

        Please overwrite this method in your Checker. A Checkers 
        purpose is to identify environment in which we choose
        different runstates for the koil. 

        Your checker should either return None if it does identify
        a certain set of conditions for a certain desired state or
        should return the desired state

        Raises:
            CheckerExpection: An exception if there is a mismatch between a nesecarry state and the current environment we are in (e.g no desired library)

        Returns:
            Union[None, KoilState]: The derised State or None if no state is desired
        """



        raise CheckerExpection("Checker did not overwrite the proper function")




import threading
import asyncio 
from asyncio.runners import _cancel_all_tasks
from threading import Thread
import os
import logging
import time

from koil.checker.registry import get_checker_registry
from koil.state import KoilState

logger = logging.getLogger(__name__)


def newloop(loop, loop_started):
    asyncio.set_event_loop(loop)
    try:
        loop_started.set()
        logger.info("Running New Event loop in another Thread")
        loop.run_forever()
    finally:
        logger.info("Loop Shutting Down")
        try:
            _cancel_all_tasks(loop)
            loop.run_until_complete(loop.shutdown_asyncgens())
        finally:
            asyncio.set_event_loop(None)
            loop.close()


class Koil:

    def __init__(self,
        force_sync = False,
        force_async = False,
        register_default_checkers = True,
        **overrides,
    ) -> None:
        """ Creates A Herre Client

        Args:
            config_path (str, optional): [description]. Defaults to "bergen.yaml".
            username (str, optional): [description]. Defaults to None.
            password (str, optional): [description]. Defaults to None.
            allow_insecure (bool, optional): [description]. Defaults to False.
            in_sync (bool, optional): Should we force an in_sync modus if an event loop is already running. Loop will be send to another thread. Defaults to True.

        Raises:
            HerreError: [description]
        """

        self.loop = None
        self.thread_id = None
        self.state = get_checker_registry(register_defaults=register_default_checkers).get_desired_state(self)
        
        if force_sync or force_async:
            self.state.threaded = force_sync and not force_async # Force async has priority



        if self.state.threaded:
            self.loop = asyncio.new_event_loop()
            self.loop_started_event = threading.Event()
            self.thread = Thread(target=newloop, args=(self.loop,self.loop_started_event))
            self.thread.start()
            self.loop_started_event.wait()
            logger.info("Running in Seperate Thread so that we can use the sync syntax")
        else:
            try:
                self.loop = asyncio.get_running_loop()
            except RuntimeError as e:
                self.loop = asyncio.new_event_loop()
                asyncio.set_event_loop(self.loop)

        set_current_koil(self)


    async def aclose(self):
        loop = asyncio.get_event_loop()

    def close(self):
        # Do according to state
        if self.state.threaded:
            self.loop.call_soon_threadsafe(self.loop.stop())

            while self.loop.is_running():
                logger.info("Waiting for the Loop to close")
                time.sleep(0.1)



class KoiledContext():

    def __init__(self) -> None:
        pass

    def __enter__(self):
        self.koil = Koil(force_sync=True)
        return self.koil

    def __exit__(self,*args, **kwargs):
        self.koil.close()
        return

    async def __aenter__(self):
        self.koil = Koil(force_async=True)
        return self.koil

    async def __aexit__(self,*args, **kwargs):
        await self.koil.aclose()
        self.koil = None


koiled = KoiledContext()

CURRENT_KOIL = None

def get_current_koil(**kwargs):
    global CURRENT_KOIL
    if not CURRENT_KOIL:
        CURRENT_KOIL = Koil(**kwargs)
    return CURRENT_KOIL

def set_current_koil(koil):
    global CURRENT_KOIL
    CURRENT_KOIL = koil


from typing import Type, Union
from koil.checker.base import BaseChecker
from koil.checker.registry import register_checker
from koil.state import KoilState
from koil.task import KoilTask


class JupyterKoilState(KoilState):
    
    def __init__(self, is_terminal=False, **kwargs) -> None:
        super().__init__(**kwargs)
        self.is_terminal = is_terminal

    def get_task_class(self) -> Type[KoilTask]:
        return KoilTask
        


@register_checker()
class JupyterChecker(BaseChecker):

    def force_state(self) -> Union[None, KoilState]:
        """Checks if a running Qt Instance is there, if so we would like
        to run in a seperate thread

        Returns:
            Union[None, KoilState]: [description]
        """

        try:
            from IPython import get_ipython
            shell = get_ipython().__class__.__name__
            if shell == 'ZMQInteractiveShell':
                return JupyterKoilState(threaded=True)   # Jupyter notebook or qtconsole
            elif shell == 'TerminalInteractiveShell':
                return JupyterKoilState(threaded=True, is_terminal=True)   # Terminal running IPython
            else:
                return None  # Other type (?)
        except NameError:
            return None  
        except ImportError:
            return None





import asyncio
import threading
import logging


logger = logging.getLogger(__name__)


def unkoil(function):

    def koiled_function(*args, **kwargs):
        return koil(function(*args, **kwargs))

    return koiled_function


def koil(potential_future, as_task=False, block=False, koil= None):
    """Taks an Async Function and according
    to the current loop setting either returns over
    it synchronusly in the herre loop or returns
    a future

    Args:
        potential_future ([type]): [description]

    Returns:
        [type]: [description]
    """
    from koil.koil import get_current_koil

    koil = koil or get_current_koil()
    state = koil.state
    loop = koil.loop


    if loop.is_running():
        # Loop is running
        logger.debug("Loop is Running")
        if loop._thread_id == threading.current_thread().ident: 
            logger.debug(f"Same thread... Returning Future {potential_future}")
            # Loop is running in same thread, always returning future here (in an ansyc context you are the master)
            return potential_future
        # We are in a threaded context our state is importat
        if not block and as_task:
            return koil.state.get_task_class()(future=potential_future, koil=koil)

        return asyncio.run_coroutine_threadsafe(potential_future, loop).result()
    
    # loop is not running and we have no threaded loop, just returning the futures

    return loop.run_until_complete(potential_future)



def next_on_gen(potential_generator, loop):
    """Takes a Async Generator and iterates over it
    threadsafe in the provided loop and returns the result
    synchronously in another generator

    Args:
        potential_generator ([type]): [description]
        loop ([type]): [description]

    Returns:
        [type]: [description]

    Yields:
        [type]: [description]
    """
    ait = potential_generator.__aiter__()

    async def next_on_ait():
        try:
            obj = await ait.__anext__()
            return False, obj
        except StopAsyncIteration:
            return True, None
        
    while True:
        done, obj = asyncio.run_coroutine_threadsafe(next_on_ait(), loop).result()
        if done: 
            break
        yield obj

def next_on_gen_not_running(potential_generator, loop):
    """Takes a Async Generator and iterates over it
    threadsafe in the provided loop and returns the result
    synchronously in another generator

    Args:
        potential_generator ([type]): [description]
        loop ([type]): [description]

    Returns:
        [type]: [description]

    Yields:
        [type]: [description]
    """
    ait = potential_generator.__aiter__()

    async def next_on_ait():
        try:
            obj = await ait.__anext__()
            return False, obj
        except StopAsyncIteration:
            return True, None
        
    while True:
        done, obj = loop.run_until_complete(next_on_ait())
        if done: 
            break
        yield obj


def koil_gen(potential_generator, koil = None):
    """Takes a Async Generator and iterates over it
    threadsafe in the herre loop providing a syncrhonus
    generator or returns an async generator in the current
    loop

    Args:
        potential_generator ([type]): [description]

    Returns:
        [type]: [description]
    """
    from koil.koil import get_current_koil
    
    koil = koil or get_current_koil()
    loop = koil.loop

    if loop.is_running():
        if loop._thread_id == threading.current_thread().ident: return potential_generator
        return next_on_gen(potential_generator, loop)


    return next_on_gen_not_running(potential_generator, loop)



